/*
 * Copyright (C) 2011 Filippo Forchino <filippo.forchino@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "SmartCam.h"
#include "gui/MainMenuForm.h"
#include "gui/CameraForm.h"
#include "gui/SettingsForm.h"
#include "gui/HelpForm.h"
#include "gui/AboutForm.h"
#include "video/JpegHandler.h"
#include "settings/UserSettings.h"

using namespace Osp::App;
using namespace Osp::Base;
using namespace Osp::System;
using namespace Osp::Ui;
using namespace Osp::Ui::Controls;
using namespace Osp::Graphics;

SmartCam::SmartCam():
	pCamera(NULL),
	isMainCameraAvailable(false),
	isFrontCameraAvailable(false),
	// MAIN camera
	pMainPreviewFormatList(NULL),
	pMainPreviewResolutionList(NULL),
	pMainPreviewFPSList(NULL),
	// FRONT camera
	pFrontPreviewFormatList(NULL),
	pFrontPreviewResolutionList(NULL),
	pFrontPreviewFPSList(NULL),
	// other handlers
	pCommHandler(NULL),
	pJpegHandler(NULL),
	pSettings(NULL),
	// UI forms
	pCameraForm(NULL),
	pSettingsForm(NULL),
	pHelpForm(NULL),
	pAboutForm(NULL),
	pMainMenuForm(NULL),
	initErrMsg(""),
	isTerminating(false)
{
}

void SmartCam::dispose()
{
	if(pJpegHandler != NULL)
	{
		delete pJpegHandler;
		pJpegHandler = NULL;
	}
	// MAIN camera
	if(pMainPreviewFormatList != NULL)
	{
		pMainPreviewFormatList->RemoveAll();
		delete pMainPreviewFormatList;
		pMainPreviewFormatList = NULL;
	}
	if(pMainPreviewResolutionList != NULL)
	{
		pMainPreviewResolutionList->RemoveAll(true);
		delete pMainPreviewResolutionList;
		pMainPreviewResolutionList = NULL;
	}
	if(pMainPreviewFPSList != NULL)
	{
		delete pMainPreviewFPSList;
		pMainPreviewFPSList = NULL;
	}
	// FRONT camera
	if(pFrontPreviewFormatList != NULL)
	{
		pFrontPreviewFormatList->RemoveAll();
		delete pFrontPreviewFormatList;
		pFrontPreviewFormatList = NULL;
	}
	if(pFrontPreviewResolutionList != NULL)
	{
		pFrontPreviewResolutionList->RemoveAll(true);
		delete pFrontPreviewResolutionList;
		pFrontPreviewResolutionList = NULL;
	}
	if(pFrontPreviewFPSList != NULL)
	{
		delete pFrontPreviewFPSList;
		pFrontPreviewFPSList = NULL;
	}

	if(pCamera != NULL)
	{
		delete pCamera;
		pCamera = NULL;
	}
	if(pSettings != NULL)
	{
		delete pSettings;
		pSettings = NULL;
	}
	if (pConnThread)
	{
		pConnThread->Stop();
		delete pConnThread;
	}
	if (bmpBGP) delete bmpBGP;
	if (bmpBGL) delete bmpBGL;

	AppLog("Application Ends Successfully");
}

SmartCam::~SmartCam()
{
	AppLog("~SmartCam START");
	//dispose();
	AppLog("~SmartCam END");
}

void SmartCam::DisplayMainMenuForm()
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pMainMenuForm);

	// Draw and Show the form
	pMainMenuForm->Draw();
	pMainMenuForm->Show();
	Osp::System::PowerManager::KeepScreenOnState(false, true);
}
void SmartCam::DisplayCameraForm()
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pCameraForm);

	// Draw and Show the form
	pCameraForm->Draw();
	pCameraForm->Show();
	Osp::System::PowerManager::KeepScreenOnState(true, false);
}

void SmartCam::DisplaySettingsForm()
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pSettingsForm);

	pSettingsForm->PopulateFromSettings();
	// Draw and Show the form
	pSettingsForm->Draw();
	pSettingsForm->Show();
}

void SmartCam::DisplayHelpForm()
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pHelpForm);

	// Draw and Show the form
	pHelpForm->Draw();
	pHelpForm->Show();
}

void SmartCam::DisplayAboutForm()
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pAboutForm);

	// Draw and Show the form
	pAboutForm->Draw();
	pAboutForm->Show();
}

void SmartCam::ShowErrorMsg(const String& errMsg)
{
    MessageBox messageBox;
    messageBox.Construct(L"Error", errMsg, MSGBOX_STYLE_OK);

    // Call ShowAndWait - draw, show itself and process events
    int modalResult = 0;
    messageBox.ShowAndWait(modalResult);
}

void SmartCam::ShowErrorMsgAndExit(const String& errMsg)
{
    MessageBox messageBox;
    messageBox.Construct(L"Error", errMsg, MSGBOX_STYLE_OK);

    // Call ShowAndWait - draw, show itself and process events
    int modalResult = 0;
    messageBox.ShowAndWait(modalResult);

    Terminate();
}

void SmartCam::ShowForm(Osp::Ui::Controls::Form* pForm)
{
	Frame* pFrame = GetAppFrame()->GetFrame();
	// Set the current form
	pFrame->SetCurrentForm(*pAboutForm);

	// Draw and Show the form
	pForm->Draw();
	pForm->Show();
}

Application* SmartCam::CreateInstance(void)
{
	// Create the instance through the constructor.
	return new SmartCam();
}

bool SmartCam::OnAppInitializing(AppRegistry& appRegistry)
{
	AppLog( "SmartCam::OnAppInitializing()");
	// Load settings
	pSettings = UserSettings::LoadSettings();

	bmpBGP = Osp::App::Application::GetInstance()->GetAppResource()->GetBitmapN(L"bgP.png", BITMAP_PIXEL_FORMAT_ARGB8888);
	bmpBGL = Osp::App::Application::GetInstance()->GetAppResource()->GetBitmapN(L"bgL.png", BITMAP_PIXEL_FORMAT_ARGB8888);

	// Get MAIN camera caps
	CameraHandler* pMainCamera = new CameraHandler(*this, Osp::Media::CAMERA_PRIMARY);
	if(pMainCamera == NULL)
	{
		AppLogException("SmartCam::OnAppInitializing(): new CameraHandler() returned NULL for main camera!");
		return false;
	}

	result r = pMainCamera->Initialize();
	if(IsFailed(r))
	{
		AppLogException("SmartCam::OnAppInitializing(): pMainCamera->Initialize() failed (%d): [%s]", r, GetErrorMessage(r));
		isMainCameraAvailable = false;
		pMainPreviewFormatList = NULL;
		pMainPreviewResolutionList = NULL;
		pMainPreviewFPSList = NULL;
	}
	else
	{
		isMainCameraAvailable = true;
		pMainPreviewFormatList = pMainCamera->GetSupportedPreviewFormatListN();
		pMainPreviewResolutionList = pMainCamera->GetSupportedPreviewResolutionListN();
		pMainPreviewFPSList = pMainCamera->GetSupportedPreviewFrameRateListN();
	}
	delete pMainCamera;

	// Get FRONT camera caps
	CameraHandler* pFrontCamera = new CameraHandler(*this, Osp::Media::CAMERA_SECONDARY);
	if(pFrontCamera == NULL)
	{
		AppLogException("SmartCam::OnAppInitializing(): new CameraHandler() returned NULL for front camera!");
		delete pMainCamera;
		return false;
	}

	r = pFrontCamera->Initialize();
	if(IsFailed(r))
	{
		AppLogException("SmartCam::OnAppInitializing(): pFrontCamera->Initialize() failed (%d): [%s]", r, GetErrorMessage(r));
		isFrontCameraAvailable = false;
		pFrontPreviewFormatList = NULL;
		pFrontPreviewResolutionList = NULL;
		pFrontPreviewFPSList = NULL;
	}
	else
	{
		isFrontCameraAvailable = true;
		pFrontPreviewFormatList = pFrontCamera->GetSupportedPreviewFormatListN();
		pFrontPreviewResolutionList = pFrontCamera->GetSupportedPreviewResolutionListN();
		pFrontPreviewFPSList = pFrontCamera->GetSupportedPreviewFrameRateListN();
	}
	delete pFrontCamera;

	if(!isMainCameraAvailable && !isFrontCameraAvailable)
	{
		AppLogException("SmartCam::OnAppInitializing(): No camera is available, exiting !");
		initErrMsg = GS(L"IDS_NO_CAMERA_AVAILABLE");
		return false;
	}

	pConnThread = new ConnectionThread(this);
	pConnThread->Init();

	pCommHandler = pConnThread->pCommHandler;
	if(null==pCommHandler)
	{
		AppLogException("SmartCam::OnAppInitializing(): CommHandler::Construct() failed, error=%d", r);
		initErrMsg = GS(L"IDS_COULD_NOT_START_NETWORK_CONNECTION");
		return false;
	}
	pConnThread->Start();

	// Create the app forms
	pCameraForm = new CameraForm();
	pCameraForm->Initialize();

	pSettingsForm = new SettingsForm();
	pSettingsForm->Initialize();

	pHelpForm = new HelpForm();
	pHelpForm->Initialize();

	pAboutForm = new AboutForm();
	pAboutForm->Initialize();

	pMainMenuForm = new MainMenuForm();
	pMainMenuForm->Initialize();

	// Add the forms to the frame
	Frame *pFrame = GetAppFrame()->GetFrame();
	pFrame->AddControl(*pCameraForm);
	pFrame->AddControl(*pSettingsForm);
	pFrame->AddControl(*pHelpForm);
	pFrame->AddControl(*pAboutForm);
	pFrame->AddControl(*pMainMenuForm);

	DisplayMainMenuForm();

	return true;
}

bool SmartCam::OnAppTerminating(AppRegistry& appRegistry, bool forcedTermination)
{
	AppLog("SmartCam::OnAppTerminating()");
	isTerminating = true;

	SaveSettings(pSettings);

	if(!initErrMsg.IsEmpty())
		ShowErrorMsg(initErrMsg);
	dispose();

	return true;
}

void SmartCam::OnForeground(void)
{
	AppLog("SmartCam::OnForeground()");
}

void SmartCam::OnBackground(void)
{
	AppLog("SmartCam::OnBackground()");
}

void SmartCam::OnLowMemory(void)
{
	AppLog("SmartCam::OnLowMemory()");
}

void SmartCam::OnBatteryLevelChanged(BatteryLevel batteryLevel)
{
	AppLog("SmartCam::OnBatteryLevelChanged()");
}

void SmartCam::OnScreenOn()
{
	AppLog("SmartCam::OnScreenOn()");
}

void SmartCam::OnScreenOff()
{
	AppLog("SmartCam::OnScreenOff()");
}

bool SmartCam::IsConnected()
{
	return (pCommHandler->GetCommState() == CommHandler::Connected);
}

// From CommHandlerObserver:
void SmartCam::OnConnected()
{
	AppLog("SmartCam::OnConnected()");

    String errMsg = "";
	result r = E_SUCCESS;
	Osp::Ui::Orientation orientation = ((Osp::Media::CameraSelection)pSettings->cameraSource) == Osp::Media::CAMERA_PRIMARY ? Osp::Ui::ORIENTATION_LANDSCAPE : Osp::Ui::ORIENTATION_PORTRAIT;
	pCamera = new CameraHandler(*this, (Osp::Media::CameraSelection)pSettings->cameraSource);
	if(pCamera == NULL)
	{
		AppLogException("SmartCam::OnConnected(): new CameraHandler() returned NULL for camera %d!", pSettings->cameraSource);
		errMsg = "Error opening camera, exiting !";
		goto CATCH;
	}

	r = pCamera->Initialize();
	if(IsFailed(r))
	{
		delete pCamera;
		pCamera = NULL;
		errMsg = "Can not open camera, close any applications using the camera(s) and try again.";
		goto CATCH;
	}

	DisplayCameraForm();
	pCameraForm->EnableCameraPreview(orientation);
	r = pCamera->StartPreview(pCameraForm->GetBufferInfo());
	if(IsFailed(r))
	{
		delete pCamera;
		pCamera = NULL;
		errMsg = "Can not open camera, close any applications using the camera(s) and try again.";
		goto CATCH;
	}

	if(pJpegHandler != NULL)
	{
		delete pJpegHandler;
	}
	pJpegHandler = new JpegHandler();

	switch(pCamera->GetPreviewFormat())
	{
	case Osp::Graphics::PIXEL_FORMAT_YCbCr420_PLANAR:
		pJpegHandler->InitYUV();
		//if(pCommHandler->GetCommState() == CommHandler::ConnectedReadyToSend)
		{
			pCommHandler->SendData(pJpegHandler->GetOutputBuffer());
		}
		break;
	default:
		errMsg = "SmartCam does not support the camera format on your phone";
		goto CATCH;
	}

CATCH:
	pMainMenuForm->HideWaitDlg();
	if(!errMsg.IsEmpty())
	{
		ShowErrorMsgAndExit(errMsg);
	}
}

void SmartCam::OnConnectError(result errCode, const String& errMsg)
{
	pMainMenuForm->HideWaitDlg();
	ShowErrorMsg(errMsg);
	pCommHandler->Disconnect();
	pCameraForm->DisableCameraPreview();
}

void SmartCam::OnDisconnected()
{
	pMainMenuForm->HideWaitDlg();
	pCameraForm->DisableCameraPreview();
	delete pCamera;
	pCamera = NULL;
	if (!isTerminating) DisplayMainMenuForm();
}

void SmartCam::OnFrameAvailable(Osp::Base::ByteBuffer& frameData)
{
	ByteBuffer copyBuf;
	copyBuf.Construct(frameData.GetCapacity());
	copyBuf.CopyFrom(frameData);
	if(pCommHandler->GetCommState() == CommHandler::Connected)
	{
		Osp::Graphics::Dimension res = GetPreviewResolution();
		pJpegHandler->EncodeYUV420P(copyBuf, res.width, res.height);
		//pJpegHandler->EncodeYUV420P(frameData, res.width, res.height);
		pCommHandler->SendData(pJpegHandler->GetOutputBuffer());
	}
}

Osp::Graphics::Dimension SmartCam::GetPreviewResolution()
{
	return pCamera->GetPreviewResolution();
}

// MAIN camera
Osp::Base::Collection::IListT<Osp::Graphics::PixelFormat>* SmartCam::GetMainPreviewFormatList()
{
	return pMainPreviewFormatList;
}

Osp::Base::Collection::IList* SmartCam::GetMainPreviewResolutionList()
{
	return pMainPreviewResolutionList;
}

Osp::Base::Collection::IListT<int>* SmartCam::GetMainPreviewFPSList()
{
	return pMainPreviewFPSList;
}

// FRONT camera
Osp::Base::Collection::IListT<Osp::Graphics::PixelFormat>* SmartCam::GetFrontPreviewFormatList()
{
	return pFrontPreviewFormatList;
}

Osp::Base::Collection::IList* SmartCam::GetFrontPreviewResolutionList()
{
	return pFrontPreviewResolutionList;
}

Osp::Base::Collection::IListT<int>* SmartCam::GetFrontPreviewFPSList()
{
	return pFrontPreviewFPSList;
}

bool SmartCam::IsMainCameraAvailable()
{
	return isMainCameraAvailable;
}

bool SmartCam::IsFrontCameraAvailable()
{
	return isFrontCameraAvailable;
}

void SmartCam::Connect()
{
	pMainMenuForm->ShowWaitDlg();
	//pCommHandler->Connect(pSettings->serverAddr, pSettings->serverPort);
	pConnThread->SendUserEvent(ConnectionThread::OPERATION_CONNECT, null);
}

void SmartCam::Disconnect()
{
	//pCommHandler->Disconnect();
	pConnThread->SendUserEvent(ConnectionThread::OPERATION_DISCONNECT, null);
}

UserSettings& SmartCam::GetSettings()
{
	return *pSettings;
}

void SmartCam::SaveSettings(UserSettings* pNewSettings)
{
	if((pSettings != pNewSettings) && (*pSettings != *pNewSettings))
	{
        UserSettings::SaveSettings(pNewSettings);
		*pSettings = *pNewSettings;
	}
}

void SmartCam::OnUserEventReceivedN(RequestId requestId, Osp::Base::Collection::IList* pArgs)
{
	AppLog("SmartCam received event %d", requestId);
	switch(requestId)
	{
		case SmartCam::EVENT_ON_CONNECTED:
		{
			this->OnConnected();
		}
		break;
		case SmartCam::EVENT_ON_DISCONNECTED:
		{
			this->OnDisconnected();
		}
		break;
		case SmartCam::EVENT_ON_CONNECTION_ERROR:
		{
			// Process the event arguments
			if (pArgs!=null)
			{
				Integer *pErrCode = (Integer *)pArgs->GetAt(0);
				String *pErrMsg = (String *)pArgs->GetAt(1);

				this->OnConnectError(pErrCode->ToInt(),*pErrMsg);
				delete pArgs;
				if (pErrCode) delete pErrCode;
				if (pErrMsg) delete pErrMsg;
			}
		}
		default:
		{
		}
		break;
	}
}

String SmartCam::GS(String resId)
{
	String trans;
	SmartCam *pApp = static_cast<SmartCam*>(Osp::App::Application::GetInstance());
	pApp->GetAppResource()->GetString(resId, trans);
	if (trans.IsEmpty()) trans = resId;
	return trans;
}



