/*
 * Copyright (C) 2011 Filippo Forchino <filippo.forchino@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "SmartCam.h"
#include "comm/CommHandler.h"

CommHandler::CommHandler(CommHandlerObserver& commObserver):
	pSocket(NULL),
	pDns(NULL),
	port(0),
	commState(Disconnected),
	observer(commObserver)
{
	pSocket = new Socket();
}

CommHandler::~CommHandler()
{
	Disconnect();
	if(pDns != NULL)
	{
		delete pDns;
		pDns = NULL;
	}
	delete pSocket;
	pSocket = NULL;
}

result CommHandler::Construct()
{
	result r = E_SUCCESS;
	// Construct the dns resolver
	pDns = new Dns();
	r = pDns->Construct(*this);
	if(IsFailed(r))
	{
		AppLogException("CommHandler::Construct(): Dns::Construct() failed, error=%d: %s", r, GetErrorMessage(r));
	}
	return r;
}

result CommHandler::CreateSocket()
{
	result r = E_SUCCESS;
	pSocket = new Socket();
	// Construct TCP stream socket
	r = pSocket->Construct(NET_SOCKET_AF_IPV4, NET_SOCKET_TYPE_STREAM, NET_SOCKET_PROTOCOL_TCP);
	if(IsFailed(r))
	{
		AppLog("CommHandler::CreateSocket(): Socket::Construct() failed, error=%d: %s", r, GetErrorMessage(r));
		return r;
	}

	unsigned long value = 0;
	r = pSocket->Ioctl(NET_SOCKET_FIONBIO, value);
	if(IsFailed(r))
	{
		AppLog("Socket::CreateSocket() Socket::Ioctl() failed, error=%d: %s", r, GetErrorMessage(r));
		return r;
	}

	r = pSocket->SetSockOpt(NET_SOCKET_IPPROTO_TCP, NET_SOCKET_TCP_NODELAY, 1);
	if(IsFailed(r))
	{
		AppLog("Socket::CreateSocket() Socket::SetSockOpt() failed, error=%d: %s", r, GetErrorMessage(r));
		return r;
	}

	return r;
}

result CommHandler::Connect(String addr, int portNo)
{
	port = portNo;
	result r = CreateSocket();
	if(IsFailed(r))
	{
		observer.OnConnectError(r, GS(L"IDS_CONNECTION_ERROR"));
		return r;
	}

	commState = Connecting;
	// Create peer endpoint
	ClearLastResult();
	Ip4Address peerAddr(addr); //
	r = GetLastResult();
	AppLog("CommHandler::Connect() - last result after Ip4Address constructor %d", r);
	if(IsFailed(r)) // error, maybe given addr is a host name, resolve it to an ip addr
	{
		r = pDns->GetHostByName(addr);
		if(IsFailed(r))
		{
			observer.OnConnectError(r, GS(L"IDS_CONNECTION_ERROR"));
		}
	}
	else
	{
		r = ConnectIPAddr(peerAddr);
	}
	return r;
}

result CommHandler::ConnectIPAddr(const IpAddress& ipAddr)
{
	result r = E_SUCCESS;
	unsigned short peerPort = port;
	Osp::Net::NetEndPoint endPoint(ipAddr, peerPort);
	// Connect the socket
	r = pSocket->Connect(endPoint);
	if(IsFailed(r))
	{
		AppLog("Socket::Conect() failed, error=%d: %s", r, GetErrorMessage(r));
		observer.OnConnectError(r, GS(L"IDS_CONNECTION_ERROR"));
	}
	else
	{
		commState = Connected;
		observer.OnConnected();
	}
	return r;
}

result CommHandler::Disconnect()
{
	result r = E_SUCCESS;
	if(pSocket == NULL)
	{
		AppLogException("CommHandler::Disconnect(): NULL pSocket");
		return r;
	}
	r = pSocket->Close();
	if(IsFailed(r))
	{
		AppLogException("Socket::Close() failed, error=%d", (int) r);
	}
	commState = Disconnected;
	delete pSocket;
	pSocket = NULL;
	observer.OnDisconnected();
	return r;
}

result CommHandler::SendData(ByteBuffer* jpegData)
{
	result r = E_SUCCESS;
	while(jpegData->GetPosition() < jpegData->GetLimit())
	{
		r = pSocket->Send(*jpegData);
		if(IsFailed(r))
		{
			AppLogException("CommHandler::SendData(): Socket::Send() failed(%d):%s!", r, GetErrorMessage(r));
			//Disconnect();
			return r;
		}
	}
	return r;
}

CommHandler::CommState CommHandler::GetCommState()
{
	return commState;
}


void CommHandler::OnDnsResolutionCompletedN(IpHostEntry* pIpHostEntry, result r)
{
	AppLog("CommHandler::OnDnsResolutionCompletedN()");
	Ip4Address* pIp4Address = NULL;
    if(pIpHostEntry == NULL || IsFailed(r))
    {
        AppLog("CommHandler::OnDnsResolutionCompletedN(): error (%d): %s", r, GetErrorMessage(r));
        observer.OnConnectError(r, GS(L"IDS_COULD_NOT_RESOLVE_HOST_NAME"));
        delete pIpHostEntry;
        return;
    }
	Osp::Base::Collection::IList* addressList = pIpHostEntry->GetAddressList();
	if(addressList != NULL)
	{
        int count = addressList->GetCount();
        for(int i = 0; i < count; i++)
        {
            IpAddress* pIpAddress = (IpAddress*)(addressList->GetAt(i));
            if(pIpAddress->GetNetAddressFamily() == NET_AF_IPV4)
            {
            	AppLog("CommHandler::OnDnsResolutionCompletedN(): found IPv4 address: %S", pIpAddress->ToString().GetPointer());
            	pIp4Address = static_cast<Ip4Address*>(pIpAddress);
            	break;
            }
        }
	}
	if(pIp4Address == NULL) // No IPv4 address was found
	{
		delete pIpHostEntry;
		observer.OnConnectError(r, GS(L"IDS_COULD_NOT_RESOLVE_HOST_NAME"));
	}
	else
	{
		Ip4Address ip4CopyAddr(*pIp4Address);
		delete pIpHostEntry;
		ConnectIPAddr(ip4CopyAddr);
	}
}

String CommHandler::GS(String resId)
{
	String trans;
	SmartCam *pApp = static_cast<SmartCam*>(Osp::App::Application::GetInstance());
	pApp->GetAppResource()->GetString(resId, trans);
	if (trans.IsEmpty()) trans = resId;
	return trans;
}

