/*
 * Copyright (C) 2011 Filippo Forchino <filippo.forchino@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "settings/UserSettings.h"

#include <FApp.h>
#include <FMedia.h>

using namespace Osp::Base;

#define DEFAULT_INET_PORT 9361
#define	CAMERA_SOURCE_KEY "SMARTCAM_CAMERA_SOURCE"
#define MAIN_CAMERA_RESOLUTION_IDX_KEY "SMARTCAM_MAIN_CAMERA_RESOLUTION_IDX"
#define MAIN_CAMERA_FPS_IDX_KEY "SMARTCAM_MAIN_CAMERA_FPS_IDX"
#define FRONT_CAMERA_RESOLUTION_IDX_KEY "SMARTCAM_FRONT_CAMERA_RESOLUTION_IDX"
#define FRONT_CAMERA_FPS_IDX_KEY "SMARTCAM_FRONT_CAMERA_FPS_IDX"
#define SERVER_ADDR_KEY "SMARTCAM_SERVER_ADDR"
#define SERVER_PORT_KEY "SMARTCAM_SERVER_PORT"

UserSettings::UserSettings():
	cameraSource(Osp::Media::CAMERA_PRIMARY),
	mainCameraResolutionIdx(0),
	mainCameraFPSIdx(0),
	frontCameraResolutionIdx(0),
	frontCameraFPSIdx(0),
	serverAddr(" "),
	serverPort(DEFAULT_INET_PORT)
{
}

UserSettings::UserSettings(const UserSettings& settings):
	cameraSource(settings.cameraSource),
	mainCameraResolutionIdx(settings.mainCameraResolutionIdx),
	mainCameraFPSIdx(settings.mainCameraFPSIdx),
	frontCameraResolutionIdx(settings.frontCameraResolutionIdx),
	frontCameraFPSIdx(settings.frontCameraFPSIdx),
	serverAddr(settings.serverAddr),
	serverPort(settings.serverPort)
{
}

UserSettings& UserSettings::operator=(const UserSettings& settings)
{
    if(this != &settings)
    {
    	cameraSource = settings.cameraSource;
    	mainCameraResolutionIdx = settings.mainCameraResolutionIdx;
    	mainCameraFPSIdx = settings.mainCameraFPSIdx;
    	frontCameraResolutionIdx = settings.frontCameraResolutionIdx;
    	frontCameraFPSIdx = settings.frontCameraFPSIdx;
    	serverAddr = settings.serverAddr;
    	serverPort = settings.serverPort;
    }
    return *this;
}

bool UserSettings::operator==(const UserSettings& rhs) const
{
	return
		(cameraSource == rhs.cameraSource) &&
		(mainCameraResolutionIdx == rhs.mainCameraResolutionIdx) &&
		(mainCameraFPSIdx == rhs.mainCameraFPSIdx) &&
		(frontCameraResolutionIdx == rhs.frontCameraResolutionIdx) &&
		(frontCameraFPSIdx == rhs.frontCameraFPSIdx) &&
		(serverAddr == rhs.serverAddr) &&
		(serverPort == rhs.serverPort);

}

bool UserSettings::operator!=(const UserSettings& rhs) const
{
	return !(*this == rhs);
}

UserSettings::~UserSettings()
{
}

void UserSettings::GetIntVal(Osp::App::AppRegistry& appRegistry, const Osp::Base::String& key, int& val)
{
	result r = appRegistry.Get(key, val);
    if(r == E_KEY_NOT_FOUND)
    {
        r = appRegistry.Add(key, val);
        if(IsFailed(r))
        {
        	AppLog("UserSettings::GetIntVal(): error adding key %S to the registry [%s]!", key.GetPointer(), GetErrorMessage(r));
        }
    	r = appRegistry.Save();
    	if(IsFailed(r))
    	{
    		AppLog("UserSettings::GetIntVal(): error saving default settings key %S [%s]!", key.GetPointer(), GetErrorMessage(r));
    	}

    }
    else if(IsFailed(r))
    {
    	AppLog("UserSettings::GetIntVal(): error reading key %S [%s]!", key.GetPointer(), GetErrorMessage(r));
    }
}

void UserSettings::GetStringVal(Osp::App::AppRegistry& appRegistry, const Osp::Base::String& key, Osp::Base::String& val)
{
	result r = appRegistry.Get(key, val);
    if(r == E_KEY_NOT_FOUND)
    {
        r = appRegistry.Add(key, val);
        if(IsFailed(r))
        {
        	AppLog("UserSettings::GetStringVal(): error adding key %S to the registry [%s]!", key.GetPointer(), GetErrorMessage(r));
        }
    	r = appRegistry.Save();
    	if(IsFailed(r))
    	{
    		AppLog("UserSettings::GetStringVal(): error saving default settings key %S [%s]!", key.GetPointer(), GetErrorMessage(r));
    	}
    }
    else if(IsFailed(r))
    {
    	AppLog("UserSettings::GetStringVal(): error reading key %S [%s]!", key.GetPointer(), GetErrorMessage(r));
    }
}

UserSettings* UserSettings::LoadSettings()
{
	UserSettings* pSettings = new UserSettings(); // default settings constructor
	Osp::App::AppRegistry* pAppRegistry = Osp::App::Application::GetInstance()->GetAppRegistry();
	// Camera source
	GetIntVal(*pAppRegistry, CAMERA_SOURCE_KEY, pSettings->cameraSource);
	// Main camera resolution
	GetIntVal(*pAppRegistry, MAIN_CAMERA_RESOLUTION_IDX_KEY, pSettings->mainCameraResolutionIdx);
	// Main camera FPS
	GetIntVal(*pAppRegistry, MAIN_CAMERA_FPS_IDX_KEY, pSettings->mainCameraFPSIdx);
	// Front camera resolution
	GetIntVal(*pAppRegistry, FRONT_CAMERA_RESOLUTION_IDX_KEY, pSettings->frontCameraResolutionIdx);
	// Front camera FPS
	GetIntVal(*pAppRegistry, FRONT_CAMERA_FPS_IDX_KEY, pSettings->frontCameraFPSIdx);
	// Server address
	String serverAddrKey(SERVER_ADDR_KEY);
	GetStringVal(*pAppRegistry, serverAddrKey, pSettings->serverAddr);
	// Server port
	GetIntVal(*pAppRegistry, SERVER_PORT_KEY, pSettings->serverPort);
	return pSettings;
}

void UserSettings::SaveSettings(UserSettings* pSettings)
{
	result r = E_SUCCESS;
	Osp::App::AppRegistry* pAppRegistry = Osp::App::Application::GetInstance()->GetAppRegistry();
	// Camera source
	r = pAppRegistry->Set(CAMERA_SOURCE_KEY, pSettings->cameraSource);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing camera source [%s]!", GetErrorMessage(r));
	}
	// Main camera resolution
	r = pAppRegistry->Set(MAIN_CAMERA_RESOLUTION_IDX_KEY, pSettings->mainCameraResolutionIdx);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing main camera resolution [%s]!", GetErrorMessage(r));
	}
	// Main camera FPS
	r = pAppRegistry->Set(MAIN_CAMERA_FPS_IDX_KEY, pSettings->mainCameraFPSIdx);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing main camera FPS [%s]!", GetErrorMessage(r));
	}
	// Front camera resolution
	r = pAppRegistry->Set(FRONT_CAMERA_RESOLUTION_IDX_KEY, pSettings->frontCameraResolutionIdx);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing front camera resolution [%s]!", GetErrorMessage(r));
	}
	// Front camera FPS
	r = pAppRegistry->Set(FRONT_CAMERA_FPS_IDX_KEY, pSettings->frontCameraFPSIdx);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing front camera FPS [%s]!", GetErrorMessage(r));
	}
	// Server address
	r = pAppRegistry->Set(SERVER_ADDR_KEY, pSettings->serverAddr.GetPointer());
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing server address [%s]!", GetErrorMessage(r));
	}
	// Server port
	r = pAppRegistry->Set(SERVER_PORT_KEY, pSettings->serverPort);
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error writing server port [%s]!", GetErrorMessage(r));
	}

	r = pAppRegistry->Save();
	if(IsFailed(r))
	{
		AppLog("UserSettings::SaveSettings(): error saving the registry [%s]!", GetErrorMessage(r));
	}
}
