/*
 * Copyright (C) 2011 Filippo Forchino <filippo.forchino@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "video/CameraHandler.h"
#include "SmartCam.h"
#include "settings/UserSettings.h"

using namespace Osp::Base;
using namespace Osp::Media;
using namespace Osp::Base::Collection;
using namespace Osp::Graphics;

CameraHandler::CameraHandler(CameraHandlerObserver& cameraObserver, CameraSelection camSel):
	observer(cameraObserver),
	cameraSel(camSel)
{
}

CameraHandler::~CameraHandler()
{
	Terminate();
}

result CameraHandler::StartPreview(const Osp::Graphics::BufferInfo* pBufferInfo)
{
	result r = E_SUCCESS;
	CameraState state = GetState();
	AppLog("CameraHandler::StartPreview(): camera state: %d", state);
	if(state != CAMERA_STATE_INITIALIZED && state != CAMERA_STATE_CAPTURED)
	{
		r = E_INVALID_STATE;
		AppLog("CameraHandler::StartPreview(): Invalid State, state: %d", state);
		return r;
	}

	r = Camera::StartPreview(pBufferInfo, true);
	if(IsFailed(r))
	{
		AppLog("CameraHandler::StartPreview(): Camera::StartPreview() failed.");
		return r;
	}

	return r;
}

result CameraHandler::StopPreview()
{
	result r = E_SUCCESS;

	Osp::Media::CameraState state = GetState();
	AppLog("CameraHandler::StopPreview(): camera state=%d", state);

	//if(state == CAMERA_STATE_PREVIEW)
	{
		r = Camera::StopPreview();
		if(IsFailed(r))
		{
			AppLog("CameraHandler::StopPreview(): Camera::StopPreview() failed.");
			return r;
		}
	}

	return r;
}

result CameraHandler::Initialize()
{
	result r = E_SUCCESS;

	Osp::Base::Collection::IListT<PixelFormat>* pPreviewFormatList = NULL;
	PixelFormat cameraPixelFormat = PIXEL_FORMAT_MIN;
	PixelFormat crtPixelFormat = PIXEL_FORMAT_MIN;

	IList* pPreviewResolutionList = NULL;
	Dimension* pPreviewResolution = NULL;

	Osp::Base::Collection::IListT<int>* pPreviewFPSList = NULL;
	int previewFPS = 0;

	UserSettings settings = (static_cast<SmartCam*>(&observer))->GetSettings();
	int cameraResolutionIdx = 0;
	int cameraFPSIdx = 0;
	if(cameraSel == Osp::Media::CAMERA_PRIMARY)
	{
		cameraResolutionIdx = settings.mainCameraResolutionIdx;
		cameraFPSIdx = settings.mainCameraFPSIdx;
	}
	else
	{
		cameraResolutionIdx = settings.frontCameraResolutionIdx;
		cameraFPSIdx = settings.frontCameraFPSIdx;
	}
	r = Camera::Construct(*this, cameraSel);
	if(IsFailed(r))
	{
		AppLog("CameraHandler::Initialize(): Camera::Construct() failed (%d).", r);
		goto CATCH;
	}

	if(!IsPoweredOn())
	{
		r = PowerOn();
		if(IsFailed(r))
		{
			AppLog("CameraHandler::Initialize(): Power on camera failed.");
			goto CATCH;
		}
	}

	// Set preview pixel format
	pPreviewFormatList = GetSupportedPreviewFormatListN();
	if(pPreviewFormatList == NULL )
	{
		AppLog("Getting preview formats failed.");
		goto CATCH;
	}

	for(int i = 0; i < pPreviewFormatList->GetCount(); i++)
	{
		pPreviewFormatList->GetAt(i, crtPixelFormat);
		AppLog("Format[%d]: %S", i, PixelFormat2Str(crtPixelFormat).GetPointer());
	}

	if(pPreviewFormatList->Contains(PIXEL_FORMAT_YCbCr420_PLANAR))
	{
		cameraPixelFormat = PIXEL_FORMAT_YCbCr420_PLANAR;
	}
	else
	{
		pPreviewFormatList->GetAt(0, cameraPixelFormat);
	}

	r = SetPreviewFormat(cameraPixelFormat);
	if(IsFailed(r))
	{
		AppLog("Setting preview format failed - %S !", PixelFormat2Str(cameraPixelFormat).GetPointer());
		goto CATCH;
	}

 	// Set preview resolution
	pPreviewResolutionList = GetSupportedPreviewResolutionListN();
	if(pPreviewResolutionList == NULL)
	{
		AppLog("Getting preview resolutions failed.");
		goto CATCH;
	}

	for(int i = 0; i < pPreviewResolutionList->GetCount(); i++)
	{
		Dimension* pCrtPreviewResolution = (Dimension*)pPreviewResolutionList->GetAt(i);
		AppLog("Resolution[%d]: w=%d x h=%d", i, pCrtPreviewResolution->width, pCrtPreviewResolution->height);
	}
	pPreviewResolution = (Dimension*)pPreviewResolutionList->GetAt(cameraResolutionIdx);
	r = SetPreviewResolution(*pPreviewResolution);
	if(IsFailed(r))
	{
		AppLog("Setting preview resolution failed.");
		goto CATCH;
	}

 	// Set preview FPS
	pPreviewFPSList = GetSupportedPreviewFrameRateListN();
	if(pPreviewFPSList == NULL)
	{
		AppLog("Getting preview FPS failed.");
		goto CATCH;
	}

	for(int i = 0; i < pPreviewFPSList->GetCount(); i++)
	{
		int crtFPS = 0;
		pPreviewFPSList->GetAt(i, crtFPS);
		AppLog("FPS[%d]: %d fps", i, crtFPS);
	}
	pPreviewFPSList->GetAt(cameraFPSIdx, previewFPS);

	r = SetPreviewFrameRate(previewFPS);
	if(IsFailed(r))
	{
		AppLog("Setting preview FPS failed.");
		goto CATCH;
	}

CATCH:
	if(pPreviewFormatList)
	{
		pPreviewFormatList->RemoveAll();
		delete pPreviewFormatList;
	}
	if(pPreviewResolutionList)
	{
		pPreviewResolutionList->RemoveAll(true);
		delete pPreviewResolutionList;
	}
	if(pPreviewFPSList)
	{
		delete pPreviewFPSList;
	}

	return r;
}

result CameraHandler::Terminate()
{
	result r = E_SUCCESS;

	Osp::Media::CameraState state = GetState();
	AppLog("CameraHandler::Terminate(): camera state=%d", state);

	if(IsPoweredOn())
	{
		StopPreview();

		r = PowerOff();
		if(IsFailed(r))
		{
			AppLog("CameraHandler::Terminate(): power off camera failed.");
			return r;
		}
	}

	return r;
}

void CameraHandler::OnCameraAutoFocused(bool completeCondition)
{
	AppLog("CameraHandler::OnCameraAutoFocused(): completeCondition=%d", completeCondition);
}

void CameraHandler::OnCameraPreviewed(Osp::Base::ByteBuffer& previewedData , result r)
{
	if(IsFailed(r))
	{
		AppLogException("CameraHandler::OnCameraPreviewed(): error(%d): %s", r, GetErrorMessage(r));
	}
	else
	{
		observer.OnFrameAvailable(previewedData);
	}
}

void CameraHandler::OnCameraCaptured(Osp::Base::ByteBuffer& capturedData , result r)
{
	AppLog("CameraHandler::OnCameraCaptured(): result=%d", r);
}

void CameraHandler::OnCameraErrorOccurred(Osp::Media::CameraErrorReason r)
{
	AppLogException("CameraHandler::OnCameraErrorOccurred(): error(%d) %s", r, GetErrorMessage(r));
}

String CameraHandler::PixelFormat2Str(Osp::Graphics::PixelFormat pixelFormat)
{
	switch(pixelFormat)
	{
	case PIXEL_FORMAT_MIN:
		return "PIXEL_FORMAT_MIN";
	case PIXEL_FORMAT_RGB565:
		return "PIXEL_FORMAT_RGB565";
	case PIXEL_FORMAT_ARGB8888:
		return "PIXEL_FORMAT_ARGB8888";
	case PIXEL_FORMAT_R8G8B8A8:
		return "PIXEL_FORMAT_R8G8B8A8";
	case PIXEL_FORMAT_YCbCr420_PLANAR:
		return "PIXEL_FORMAT_YCbCr420_PLANAR";
	case PIXEL_FORMAT_JPEG:
		return "PIXEL_FORMAT_JPEG";
	case PIXEL_FORMAT_MAX:
		return "PIXEL_FORMAT_MAX";
	}

	return "Unknown pixel format";
}

Osp::Graphics::Dimension CameraHandler::ScaleProportionally(
		int srcWidth, int srcHeight, int boundingWidth, int boundingHeight)
{
	Osp::Graphics::Dimension result;
	double ratio = (((double) srcWidth) / srcHeight);
	int scaledWidth = (int) (ratio * boundingHeight);
	if( scaledWidth > boundingWidth)
	{
		result.width = boundingWidth;
		result.height = (int) (boundingWidth / ratio);
	}
	else
	{
		result.width = scaledWidth;
		result.height = (int) (scaledWidth / ratio);
	}
	return result;
}
